-- An tracker history is important on most tables. Provide an tracker trigger that logs to
-- a dedicated tracker table for the major relations.
--
-- This file should be generic and not depend on application roles or structures,
-- as it's being listed here:
--
--    https://wiki.postgresql.org/wiki/Audit_trigger_91plus    
--
-- This trigger was originally based on
--   http://wiki.postgresql.org/wiki/Audit_trigger
-- but has been completely rewritten.
--
-- Should really be converted into a relocatable EXTENSION, with control and upgrade files.

CREATE EXTENSION IF NOT EXISTS hstore;

<% if true %>
DROP SCHEMA IF EXISTS tracker CASCADE;
<% end %>
CREATE SCHEMA tracker;
REVOKE ALL ON SCHEMA tracker FROM public;

COMMENT ON SCHEMA tracker IS 'Out-of-table tracing';

CREATE TYPE tracker.actions AS ENUM(
  'INSERT', 'UPDATE', 'DELETE'
);

--
-- Audited data. Lots of information is available, it's just a matter of how much
-- you really want to record. See:
--
--   http://www.postgresql.org/docs/9.1/static/functions-info.html
--
-- Remember, every column you add takes up more tracker table space and slows tracker
-- inserts.
--
-- Every index you add has a big impact too, so avoid adding indexes to the
-- tracker table unless you REALLY need them. The hstore GIST indexes are
-- particularly expensive.
--
-- It is sometimes worth copying the tracker table, or a coarse subset of it that
-- you're interested in, into a temporary table where you CREATE any useful
-- indexes and do your analysis.
--
CREATE TABLE tracker.events (
  id bigserial primary key,                     -- id IS 'Unique identifier for each tracked event';
  row_data hstore,                              -- row_data IS 'Record value. For INSERT this is the new tuple. For DELETE and UPDATE it is the old tuple.';
  changed_fields hstore,                        -- changed_fields IS 'New values of fields changed by UPDATE. Null except for UPDATE events.';
  created_at TIMESTAMP WITH TIME ZONE NOT NULL, -- created_at IS 'Wall clock time at which tracked event''s trigger call occurred';
  checked_out_at TIMESTAMP WITH TIME ZONE       -- checked_out_at IS 'Wall clock time at which event is checked out for processing';
);


REVOKE ALL ON tracker.events FROM public;

/*

[TODO] - add comments on columns

*/

COMMENT ON TABLE tracker.events IS 'History of tracked actions on tracked tables, from tracker.if_modified_func()';
-- COMMENT ON COLUMN tracker.events.session_user_name IS 'Login / session user whose statement caused the tracked event';

/*

[TODO] - create indices for common search patterns

*/

-- CREATE INDEX events_table_oid_idx ON tracker.events(table_oid);
-- CREATE INDEX events_action_idx ON tracker.events(action);

/* [TODO] - change checked_out_at to default to 'infinity' */


/*

Track changes to a table at row level.

Optional parameters to trigger in CREATE TRIGGER call:

param 0: text, name of primary key column;
param 1: text[], columns to ignore in updates. Default [].

Updates to ignored cols are omitted from changed_fields.

Updates with only ignored cols changed are not inserted
into the tracker log.

Note that the user name logged is the login role for the session.
The tracker trigger cannot obtain the active role because it is reset
by the SECURITY DEFINER invocation of the tracker trigger its self.

*/

CREATE OR REPLACE FUNCTION tracker.if_modified_func() RETURNS TRIGGER AS $body$
DECLARE
  event fq_table_name;
  h_old hstore;
  h_new hstore;
  entity_pkey_name text;
BEGIN
  IF TG_WHEN <> 'AFTER' THEN
    RAISE EXCEPTION 'tracker.if_modified_func() may only run as an AFTER trigger';
  END IF;

  -- get args

  entity_pkey_name = TG_ARGV[0];

  -- fill row

  IF (TG_OP = 'UPDATE') THEN
    event.new_fields = jsonb(NEW.*);
    event.old_fields = jsonb(OLD.*);
    event.entity_id  = event.new_fields->entity_pkey_name
  ELSIF (TG_OP = 'DELETE') THEN
    event.old_fields = jsonb(OLD.*);
    event.entity_id  = event.old_fields->entity_pkey_name
  ELSIF (TG_OP = 'INSERT') THEN
    event.new_fields = jsonb(NEW.*);
    event.entity_id  = event.new_fields->entity_pkey_name
  END IF;

  INSERT INTO tracker.events VALUES (event.*);
  RETURN NULL;
END;
$body$
LANGUAGE plpgsql
SECURITY DEFINER
SET search_path = pg_catalog, public;

/*

Add tracking support to a table.

Arguments:
   target_table:     Table name, schema qualified if not on search_path
   primary_key_name: Name of primary key column
   ignored_cols:     Columns to exclude from update diffs, ignore updates that change only ignored cols.

*/

CREATE OR REPLACE FUNCTION tracker.track_table(target_table regclass, entity_pkey_name text, ignored_cols text[]) 
  RETURNS void AS $body$
DECLARE
  _q_txt text;
  _ignored_cols_snip text = '';
BEGIN
  IF array_length(ignored_cols,1) > 0 THEN
    _ignored_cols_snip = ', ' || quote_literal(ignored_cols);
  END IF;

  EXECUTE 'DROP TRIGGER IF EXISTS track_trigger_row ON ' || quote_ident(target_table::TEXT);

  _q_txt = 'CREATE TRIGGER track_trigger_row AFTER INSERT OR UPDATE OR DELETE ON ' || 
              quote_ident(target_table::TEXT) || 
              ' FOR EACH ROW EXECUTE PROCEDURE tracker.if_modified_func(' ||
              quote_literal(entity_pkey_name) || 
              _ignored_cols_snip || ');';
  EXECUTE _q_txt;
END;
$body$
language 'plpgsql';


/*

Add tracking support to the given table. No cols are ignored. (Shortcut)

*/

CREATE OR REPLACE FUNCTION tracker.track_table(target_table regclass) RETURNS void AS $body$
SELECT tracker.track_table($1, 'id', ARRAY[]::text[]);
$body$ LANGUAGE 'sql';

COMMENT ON FUNCTION tracker.track_table(regclass) IS $body$
$body$;


--
------------------------------------------------------------------------------------
--

-- SELECT tracker.track_table('public.posts');

--SELECT tracker.track_table('public.users', 'id', array['created_at','updated_at']);
-- -- SELECT tracker.track_table('public.users', ARRAY[]::text[]);
-- -- SELECT tracker.track_table('public.users');
--
-- INSERT INTO users (email, created_at, updated_at) VALUES('me@mo' || (1000 * random())::integer , NOW(), NOW());
-- UPDATE users SET username='mimi' WHERE email LIKE '%mo%';
-- UPDATE users SET username='momo' WHERE email LIKE '%mo%';
-- DELETE FROM users WHERE email LIKE '%mo%';
--
-- SELECT table_schema || '.' || table_name || '.' || entity_pkey, action, row_data, changed_fields, created_at FROM tracker.events;
